/********************************************************/
/* Copyright (C) 2002 OSS Nokalva.  All rights reserved.*/
/********************************************************/

/* THIS FILE IS PROPRIETARY MATERIAL OF OSS Nokalva, INC. AND
 * MAY BE USED ONLY BY DIRECT LICENSEES OF OSS Nokalva, INC.
 * THIS FILE MAY NOT BE DISTRIBUTED. */

/* Prepared by OSS Nokalva, Inc.  01/06/2002      */

/* Application Program: Tcont.java

   Illustrates automatic and manual encoding/decoding of a PDU with
   a nested BIT STRING that has a contents constraint imposed.

   This sample demonstrates:
   
   1) the coder's API to enable/disable automatic decoding of the contained
      values.
      
   2) how the application can get/set an encoded and decoded component of the
      containing BIT STRING.

    To run the sample perform the following steps:

    asn1pjav containing.asn -ber -uper
    cd containing
    containing.bat javac
    cd ..
    javac -g Tcont.java
    java Tcont
*/

/* This program demonstrates the contents constraint, that is supported by 
 * release 1.5 of the ASN.1/Pure Java Tools. The contents constraint
 * can be imposed on the BIT STRING or on the OCTET STRING and specifies that
 * the binary string contains the data encoded by other encoding rules. In
 * other words, the contents constraint provides the ability to construct a
 * mixed encoding, when, say, one part of the message is encoded by the Basic
 * Encoding Rules (BER) and another part is encoded by the Packed Encoding 
 * Rules (PER).
 *
 * The sample Packed-BCAS ASN.1 syntax defines the 'Card' message, that is 
 * encoded by BER and contains the nested BIT STRING with the BBCard data 
 * packed by PER unaligned. Although this definition is purely illustrative, 
 * it has some practical sense. Once one saves the collection of 'Card's
 * into the disk file, one can easily navigate in the collection because BER
 * inserts length determinants in front of the encoding of each 'Card'. On the 
 * other hand the result data set will be compact because the actual data is
 * encoded by PER, which is less verbose than BER.
 *
 * The Tcont application demonstrates how the runtime encodes and decodes the
 * messages that are coded by the mixture of encoding rules.
 *
 * Automatic coding.
 *
 * Once the contents constraint references the standard ASN.1 encoding rule 
 * (BER, DER or PER) the runtime is able to encode and decode the contained value 
 * automatically. In order to enable automatic coding of contained values, one
 * invokes the enableContainedValueEncoding/Decoding() methods of the Coder class.
 * To demonstrate the automatic coding, Tcont uses the sample 'myCard' value,
 * that is generated by the compiler in the Packed_BCAS.java class file. 
 *
 * Manual coding.
 *
 * Another alternative is to encode and decode the contained value manually.
 * Both operations require two steps. In order to encode the value of the
 * 'Card' the application constructs the value of the Card.packedData by
 * encoding the contained value (myBBCard) with PER unaligned. Then it invokes
 * the encode() method of the BERCoder to complete the encoding of the
 * outermost PDU. The manual decoding occurs in the reverse order. First, the
 * outermost PDU is decoded. Then the application extracts the bits of 
 * Card.packedData, runs the PER unaligned decoder on these bits, and stores 
 * the value decoded in the 'packedData' field.
 */
 
/* Compiler-generated classes */
import containing.*;
import containing.packed_bcas.*;

/* Universal classes from the OSS runtime library */
import com.oss.asn1.*;
import com.oss.util.HexTool;

/* Java I/O classes */
import java.io.*;

public class Tcont {

    public static void main(String[] args)
    {
	try {
	    Containing.initialize();
	} catch (Exception e) {
	    System.out.println("Initialization error: " + e);
	    System.exit(1);
	}
	Coder coder = Containing.getBERCoder();

	// Doing automatic coding of the contained value

	// Set the coder to emit trace output
	coder.enableEncoderDebugging();
	coder.enableDecoderDebugging();
	
	// Enable automatic coding of the contained value
	coder.enableContainedValueEncoding();
	coder.enableContainedValueDecoding();
	
	System.out.println("Encoding and decoding the contained value automatically ...");
	
	// Print the value we are going to encode
	System.out.println();
	System.out.println("\tUnencoded PDU:");
	System.out.println();
	System.out.print(Packed_BCAS.myCard);
	try {
	    // Encode the outermost PDU and the contained value
	    System.out.println();
	    System.out.println("\tTracing information from the BER DEFINITE-Length encoder ...");
	    System.out.println();
	    ByteArrayOutputStream sink = new ByteArrayOutputStream();
	    coder.encode(Packed_BCAS.myCard, sink);
	    byte[] encoding = sink.toByteArray();
	    // Print the encoding
	    System.out.println();
	    System.out.println("Encoded in " + encoding.length + " bytes:");
	    HexTool.printHex(encoding);
	    try {
		// Now decode the outermost PDU and the contained value
		System.out.println();
		System.out.println("\tTracing information from the BER decoder ...");
		System.out.println();
		ByteArrayInputStream source = new ByteArrayInputStream(encoding);
		Card decoded = (Card)coder.decode(source, new Card());
		// Print the value decoded
		System.out.println();
		System.out.println("\tDecoded PDU ...");
		System.out.println();
		System.out.println(decoded);
	    } catch (Exception e) {
		System.out.println("Automatic decoding failed: " + e);
	    }
	} catch (Exception e) {
	    System.out.println("Automatic encoding failed: " + e);
	}
	    
	// Now do the encoding and the decoding of the contained value 
	// manually. First, create the value of the Card
	Card anotherCard = new Card();
	    
	System.out.println("Encoding and decoding the contained value manually ...");
	    
	// Disable automatic coding of the contained value in the main coder
	coder.disableContainedValueEncoding();
	coder.disableContainedValueDecoding();
	    
	try {
	    // Encode the contained value myBBCard into the BIT STRING
	    System.out.println();
	    System.out.println("\tEncoding the contained myBBCard value into the BIT STRING ...");
	    System.out.println("\tTracing information from the PER UNALIGNED encoder ...");
	    System.out.println();
	    Coder childCoder = Containing.getPERUnalignedCoder();
	    // Set the child coder to emit trace output
	    childCoder.enableEncoderDebugging();
	    childCoder.enableDecoderDebugging();
	    ByteArrayOutputStream sink = new ByteArrayOutputStream();
	    BBCard contained = Packed_BCAS.myBBCard;
	    childCoder.encode(contained, sink);
	    // Extract the encoding and compute the exact number of bits it
	    // occupies
	    byte[] containedEncoding = sink.toByteArray();
	    int paddingBits = childCoder.getNumberOfPaddingBits();
	    int significantBits = 8*containedEncoding.length - paddingBits;
	    // Print the containing BIT STRING
	    System.out.println();
	    System.out.println("Encoded in " + significantBits + " bits:");
	    HexTool.printHex(containedEncoding);
	    // Initialize the components of the Card we are going to encode
	    anotherCard.setOrdinalID(1);
	    anotherCard.setPackedData(
		new Card.PackedData(containedEncoding, significantBits)
	    );
	    // Print the PDU with the contained value preencoded
	    System.out.println("\tUnencoded PDU ...");
	    System.out.println();
	    System.out.println(anotherCard);
	    try {
		// Encode the outermost PDU
		System.out.println("\tTracing information from the BER DEFINITE-Length encoder ...");
		System.out.println();
		sink = new ByteArrayOutputStream();
		coder.encode(anotherCard, sink);
		byte[] encoding = sink.toByteArray();
		// Print the encoding
		System.out.println();
		System.out.println("Encoded in " + encoding.length + " bytes:");
		HexTool.printHex(encoding);
		// Now decode the Card in two steps:
		try {
		    // 1) decode the outermost PDU using BER
		    System.out.println();
		    System.out.println("\tTracing information from the BER decoder ...");
		    System.out.println();
		    ByteArrayInputStream source = 
			new ByteArrayInputStream(encoding);
		    Card decoded = (Card)coder.decode(source, new Card());
		    // Print the value decoded
		    System.out.println();
		    System.out.println("\tDecoded PDU (the contained value is not decoded yet) ...");
		    System.out.println();
		    System.out.println(decoded);
		    // 2) Now decode the value, contained in the 'packedData'
		    try {
			source = new ByteArrayInputStream(
			    decoded.getPackedData().byteArrayValue()
			);
			System.out.println("\tDecoding the value, contained in the 'packedData' the BIT STRING ...");
			System.out.println("\tTracing information from the PER UNALIGNED decoder ...");
			System.out.println();
			decoded.getPackedData().setContainedValue(
			    (BBCard)childCoder.decode(source, new BBCard())
			);
			// Now print the outermost PDU with the contained value
			// decoded
			System.out.println();
			System.out.println("\tThe PDU with the contained value decoded ...");
			System.out.println();
			System.out.println(decoded);
		    } catch (Exception e) {
			System.out.println("Decoding of the contained value failed: " + e);
		    }
		} catch (Exception e) {
		    System.out.println("Decoding of the 'Card' failed: " + e);
		}
	    } catch (Exception e) {
		System.out.println("Encoding of the 'Card' failed: " + e);
	    }
	} catch (Exception e) {
	    System.out.println("Encoding of the contained value failed: " + e);
	}

	Containing.deinitialize();
    }
}